''' 
    Python based CLI Application interface control for vaunix LSW devices
    JA 02/21/2023    Initial Verison oF CLI control Interface
    (c) 2020-2023 by Vaunix Technology Corporation, all rights reserved
'''
from ctypes import *
import ctypes
import os
import time
import sys, getopt
from time import sleep

class CLI_Vaunix_LSW:
    def __init__(self, name="lswApi", port=''):
        self.vnx = cdll.LoadLibrary(os.path.join(os.path.abspath(os.path.dirname(__file__)), '/usr/lib/liblswsock.so'))
        self.vnx.fnLSW_Init()
        self.vnx.fnLSW_SetTestMode(False)
        
    def check_deviceexists(self, devip):
        self.vnx.fnLSW_InitDevice(devip)
        return self.vnx.fnLSW_CheckDeviceReady(devip)        
    
    # Get Model Name
    def get_modelname(self, devip):
        self.data = ctypes.create_string_buffer(32)
        self.vnx.fnLSW_GetModelName(devip, ctypes.byref(self.data))
        return self.data.value

    # Ger Serial Number
    def get_serial_number(self, devip):
        self.data= ctypes.c_int(0) # create c-var
        self.vnx.fnLSW_GetSerialNumber(devip, ctypes.byref(self.data))
        return self.data.value

    # Get Software Version
    def get_swversion(self, devip):
        self.data = ctypes.create_string_buffer(32)
        self.vnx.fnLSW_GetSoftwareVersion(devip, ctypes.byref(self.data))
        return self.data.value

    # Get IP mode Static/DHCP
    def get_ipmode(self, devip):
        self.data= ctypes.c_int(0) # create c-var
        self.vnx.fnLSW_GetIPMode(devip, ctypes.byref(self.data))
        return self.data.value

    # Get Ip Address
    def get_ipaddress(self, devip):
        self.data = ctypes.create_string_buffer(32)
        self.vnx.fnLSW_GetIPAddress(devip, ctypes.byref(self.data))
        return self.data.value
        
    # Get subnet mask
    def get_subnetmask(self, devip):
        self.data = ctypes.create_string_buffer(32)
        self.vnx.fnLSW_GetNetmask(devip, ctypes.byref(self.data))
        return self.data.value

    # Get gateway
    def get_gateway(self, devip):
        self.data = ctypes.create_string_buffer(32)
        self.vnx.fnLSW_GetGateway(devip, ctypes.byref(self.data))
        return self.data.value

    # Get Max Switch Devices
    def get_maxswitchdevices(self, devip):
        self.data= ctypes.c_int(0) # create c-var
        self.vnx.fnLSW_GetMaxSwitchDevices(devip, ctypes.byref(self.data))
        return (self.data.value)

    # Get Switch port data
    def get_switchrfoutput(self, devip, swindex):
        self.data= ctypes.c_int(0) # create c-var
        self.vnx.fnLSW_GetSwitchRFoutput(devip, swindex, ctypes.byref(self.data))
        return (self.data.value)

     # Set Switch port
    def set_switchrfoutput(self, devip, swindex, switchout):
        try:
            self.vnx.fnLSW_InitDevice(devip)
            self.vnx.fnLSW_SetSwitchRFoutput(devip, swindex, switchout)
            self.vnx.fnLSW_CloseDevice(devip)
        except:
            print("An exception occurred")
        return True

    def set_savesettings(self):
        try:
            self.vnx.fnLSW_InitDevice(devip)
            self.vnx.fnLSW_SaveSettings(devip)
            self.vnx.fnLSW_CloseDevice(devip)
        except:
            print("An exception occurred")
        return True

    # Main Function call   
    def main(argv):
        ipaddress = ''
        attenuation = ''
        frequency = ''
        lswobj = CLI_Vaunix_LSW()
        try:
            opts, args = getopt.getopt(argv,"hi:g:s:S:r",["iipaddr=","grswo=", "srswo=""Svst="])
        except getopt.GetoptError as err:
            print(err)
            print ('lswcli.py argument error')
            sys.exit(2)
        for opt, arg in opts:
            if opt == '-h':
                print ('lswcli.py -i <ipaddress> -g <Get swrfout> -s <Set swrfout> -S <savesetting> -r <read>')
                sys.exit()
            elif opt in ("-i", "--iipaddr"):
                ipaddress = arg.encode('utf-8')
                print ('Ip Address:', ipaddress)
            elif opt in ("-g", "--grswo"):
                swindex = int(argv[3]) 
                if(ipaddress !=''):
                    if not lswobj.check_deviceexists(ctypes.c_char_p(ipaddress)):
                        swoutput = lswobj.get_switchrfoutput(ctypes.c_char_p(ipaddress), swindex)
                        print("Switch Output(%d):%d" %(swindex,swoutput))
                    else:
                        print ("Device not exists.. Please Check!")
                else:
                    print ("Device IP missing.. Please Check!")   
            elif opt in ("-s", "--srswo"):
                swindex = int(argv[3]) 
                swoutput = int(argv[4])
                print ('swindex:', swindex)
                print ('swoutput:', swoutput)
                if(ipaddress !=''):
                    if not lswobj.check_deviceexists(ctypes.c_char_p(ipaddress)):
                        lswobj.set_switchrfoutput(ctypes.c_char_p(ipaddress), swindex, swoutput)
                    else:
                        print ("Device not exists.. Please Check!")
                else:
                    print ("Device IP missing.. Please Check!")         
            
            # Set savesettings
            elif opt in ("-S", "--Svst"):
                if(ipaddress !=''):
                    if not lswobj.check_deviceexists(ctypes.c_char_p(ipaddress)):
                        lswobj.set_savesettings(ctypes.c_char_p(ipaddress))
                    else:
                        print ("Device not exists.. Please Check!")
                else:
                    print ("Device IP missing.. Please Check!")
          
            elif opt in ("-r", "--rdata"):
                if(ipaddress !=''):
                    if not lswobj.check_deviceexists(ctypes.c_char_p(ipaddress)):
                        print("*****************Current Information of the device**********")     
                        print("Model Name:",lswobj.get_modelname(ctypes.c_char_p(ipaddress)))
                        print("Serial #:",lswobj.get_serial_number(ctypes.c_char_p(ipaddress)))
                        print("SW Version:",lswobj.get_swversion(ctypes.c_char_p(ipaddress)))
                        print("IP Mode:",lswobj.get_ipmode(ctypes.c_char_p(ipaddress)))
                        print("IP Address:",lswobj.get_ipaddress(ctypes.c_char_p(ipaddress)))
                        print("Subnet Mask:",lswobj.get_subnetmask(ctypes.c_char_p(ipaddress)))
                        print("Gateway:",lswobj.get_gateway(ctypes.c_char_p(ipaddress)))
                        print("Max Switch Devices:",lswobj.get_maxswitchdevices(ctypes.c_char_p(ipaddress)))
                        
                    else:
                        print ("Device not exists.. Please Check!")
                else:
                    print("Device IP missing.. Please Check!")
                

if __name__ == '__main__':
    CLI_Vaunix_LSW.main(sys.argv[1:])
    
